import 'package:mockor/src/annotations/experimental.dart';

///the annotation should be placed on your global mock method:
///
/// ```dart
///@GenerateMocker
/// T mock<T extends Object>({bool relaxed = false} => _$mock<T>(relaxed: relaxed)
///}
///```
class GenerateMocker {
  /// the types to generate `Mock` classes for
  final List<Type> types;

  /// if **true**, no mocks will be generated by this builder
  /// but will use the names of classes generated by Mockito.
  ///
  /// if **false**, simple one line mock classes will be generated
  /// which can be useful for `mocktail`.
  final bool useMockitoGeneratedTypes;

  /// if **true**, Mockito's own `@GenerateMocks` annotation will be added to
  /// the generated mock method. [useMockitoGeneratedTypes] must be **true** as well.
  ///
  /// if **false** but [useMockitoGeneratedTypes] is **true**,
  /// then you should include a `@GenerateMocks` to the mock method yourself.
  final bool generateMockitoAnnotation;

  /// Generate an `asMock` extension method for all [types] which casts it as
  /// the generated mock type.
  /// Due to null safety we can only use the `any` matcher on non null params when using the mocked type.
  /// Please read Mockito's [Null Safety README](https://github.com/dart-lang/mockito/blob/master/NULL_SAFETY_README.md) for more info.
  ///
  /// When using `mocktail`, should be set to `false`
  final bool generateMockExtensions;

  /// Add optional `relaxedVoid` parameter to `_$mock` method for `mocktail`.
  ///
  /// if `relaxedVoid` is `true` then return `null` for all unstubbed `void` methods
  /// and an empty `Future` for all unstubbed `Future<void>` methods.
  ///
  /// if `relaxedVoid` is `null` then it will default to `relaxed`.
  ///
  /// When using `mockito`, should be set to `false`
  @experimental
  final bool generateRelaxedVoidParameter;

  /// options for [mocktail](https://pub.dev/packages/mocktail).
  final GenerateMocktailFallbackValues? generateMocktailFallbackValues;

  /// default constructor for `mockito`
  const GenerateMocker(
    this.types, {
    this.useMockitoGeneratedTypes = true,
    this.generateMockitoAnnotation = true,
    this.generateMockExtensions = true,
    this.generateRelaxedVoidParameter = false,
    this.generateMocktailFallbackValues,
  });

  /// default constructor for [mocktail](https://pub.dev/packages/mocktail)
  const GenerateMocker.mocktail(
    this.types, {
    this.useMockitoGeneratedTypes = false,
    this.generateMockitoAnnotation = false,
    this.generateMockExtensions = false,
    this.generateRelaxedVoidParameter = false,
    this.generateMocktailFallbackValues,
  });
}

///
/// generate method with `registerFallbackValues` calls,
/// this is needed to make the `any()` function work in mocktail](https://pub.dev/packages/mocktail).
///
class GenerateMocktailFallbackValues {
  final List<Type> types;

  /// Register a mock class for all the non primitive parameters of all the methods in each class of [GenerateMocker.types]
  /// Classes with type parameters will be ignored.
  /// If a class didn't get picked up, please add it manually in [types].
  /// Find the generated method in `<filename>.fallback.dart`.
  @experimental
  final bool autoDetect;

  const GenerateMocktailFallbackValues(this.types, {this.autoDetect = false});
}
